import asyncio
import functools
import libdispatch

def dispatched_to_queue(method=None, wait=True):
    """Dispatches the method call to the object's dispatch queue

    This decorator functions assumes that the object has an attribute
    named '_queue' that contains the libdispatch.queue to dispatch the
    call to.

    Args:
        wait: If set to True the method call will be dispatched to the
            queue and but the asyncio thread will await the result
            returned from the block submitted to the dispatch queue.
            Defaults to true.

    Returns:
        The result of the method if wait is set to true. None will be
        be returned if wait is false.
    """
    def func(method):
        @functools.wraps(method)
        async def wrapper(self, *args, **kwargs):
            if wait:
                loop = asyncio.get_event_loop()
                future = loop.create_future()
                def queue_block():
                    result = method(self, *args, **kwargs)
                    def loop_block():
                        future.set_result(result)
                    loop.call_soon_threadsafe(loop_block)
                libdispatch.dispatch_async(self._queue, queue_block)
                return await future
            else:
                def queue_block():
                    method(self, *args, **kwargs)
                libdispatch.dispatch_async(self._queue, queue_block)
                return None
        return wrapper
    return func

def dispatched_to_loop(method=None):
    """Asynchronously dispatches the method call to the asyncio event loop

    This decorator assumes that the object has an attribute named 'loop'
    that contains the event loop to dispatch the call to.

    Returns:
        None
    """
    def func(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            asyncio.run_coroutine_threadsafe(functools.partial(method, self, *args, **kwargs)(), self.loop)
        return wrapper
    return func


class NSErrorException(Exception):
    def __init__(self, nserror):
        self.domain = nserror.domain()
        self.code = nserror.code()
        self.userInfo = nserror.userInfo()
        self.error = nserror

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"NSError(domain={self.domain}, code={self.code}, userInfo={self.userInfo})"



class Descriptor(object):
    """The Descriptor Base Class"""
    def __init__(self, characteristic, *args, **kwargs):
        self.characteristic = characteristic 

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Descriptor(identifier={self.identifier})"

    @property
    def identifier(self):
        """An identifier that uniquely identifies this descriptor"""
        raise NotImplementedError()


import platform

if platform.system() == "Linux":
    from aioble.bluezdbus.centralmanager import CentralManagerBlueZDbus as CentralManager
    from aioble.bluezdbus.device import DeviceBlueZDbus as Device
elif platform.system() == "Darwin":
    from aioble.corebluetooth.centralmanager import CoreBluetoothCentralManager as CentralManager
    from aioble.corebluetooth.device import CoreBluetoothDevice as Device
elif platform.system() == "Windows":
    from aioble.dotnet.centralmanager import CentralManagerDotNet as CentralManager
    from aioble.dotnet.device import DeviceDotNet as Device
else:
    from aioble.centralmanager import CentralManager
    from aioble.device import Device
    from aioble.service import Service
    from aioble.characteristic import Characteristic


import asyncio
from collections import Awaitable

from System import Action
from System.Threading.Tasks import Task

async def wrap_dotnet_task(task, loop):
    done = asyncio.Event()
    task.ContinueWith(Action[Task](lambda x: loop.call_soon_threadsafe(done.set)))
    await done.wait()
    if task.IsFaulted:
        raise Exception(task.Exception.ToString())
    return task.Result


class Service(object):
    """The Service Base Class"""
    def __init__(self, device, *args, **kwargs):
        self.device = device

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Service(identifier={self.identifier})"

    @property
    def identifier(self):
        """An identifier that uniquely identifies this service"""
        raise NotImplementedError()


import asyncio
import re
from aioble.service import Service
from aioble.bluezdbus.characteristic import CharacteristicBlueZDbus as Characteristic
import dbussy as dbus
from dbussy import \
    DBUS

_BLUEZ_DESTINATION = 'org.bluez'
_DBUS_OBJECT_MANAGER_INTERFACE = 'org.freedesktop.DBus.ObjectManager'
_GET_MANAGED_OBJECTS_METHOD = 'GetManagedObjects'

class ServiceBlueZDbus(Service):
    """The Service DotNet Class"""
    def __init__(self, device, path, uuid, loop=None):
        super(ServiceBlueZDbus, self).__init__(loop)
        self.loop = loop if loop else asyncio.get_event_loop()
        self.path = path
        self.device = device
        self.uuid = uuid
        self.characteristics = []

    async def resolve_characteristics(self):
        # Connect to system bus
        self._dbus = await dbus.Connection.bus_get_async(DBUS.BUS_SYSTEM, private=False)
        #print("Get Characteristics for {0}...".format(self.uuid))
        message = dbus.Message.new_method_call \
        (
            destination=_BLUEZ_DESTINATION,
            path="/",
            iface=_DBUS_OBJECT_MANAGER_INTERFACE,
            method=_GET_MANAGED_OBJECTS_METHOD
        )

        # Dict of {Object Path, Dict of {String, Dict of {String, Variant}}} objects)
        reply = await self._dbus.send_await_reply(message)
        values = reply.expect_return_objects("a{oa{sa{sv}}}")[0]

        characteristics_regex = re.compile(self.path + '/char[0-9abcdef]{4}$')

        self.characteristics = [Characteristic(
            service=self,
            path=objpath,
            uuid=objvalue['org.bluez.GattCharacteristic1']['UUID'][1]) for objpath, objvalue in values.items() if
            characteristics_regex.match(objpath)]

        for char in self.characteristics:
            await char.resolve_descriptors()


import asyncio

from aioble.centralmanager import CentralManager

class Device(object):
    """The Device Base Class"""
    def __init__(self, manager: CentralManager, loop=None, *args, **kwargs):
        self.manager = manager
        self.loop = loop if loop else asyncio.get_event_loop()
        self.properties = None
        self.services = {}
        self._notification_callbacks = {}

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Device(identifier={self.identifier}, name={self.name})"

    @property
    def identifier(self):
        """An identifier that uniquely identifies this device"""
        raise NotImplementedError()

    @property
    def name(self):
        """A user friendly name for the device device"""
        raise NotImplementedError()

    async def connect(self):
        """Connect to device"""
        raise NotImplementedError()

    async def disconnect(self):
        """Disconnect to device"""
        raise NotImplementedError()

    async def is_connected(self):
        """Is Connected to device"""
        raise NotImplementedError()

    async def get_properties(self):
        """Get Device Properties"""
        raise NotImplementedError()

    async def discover_services(self):
        """Discover Device Services"""
        raise NotImplementedError()

    async def read_char(self):
        """Read Service Char"""
        raise NotImplementedError()

    async def write_char(self):
        """Write Service Char"""
        raise NotImplementedError()

    async def start_notify(self):
        """Start Notification Subscription"""
        raise NotImplementedError()

    async def stop_notify(self):
        """Stop Notification Subscription"""
        raise NotImplementedError()


import asyncio

class CentralManager(object):
    """The Central Manager Base Class"""
    def __init__(self, loop=None, *args, **kwargs):
        self.loop = loop if loop else asyncio.get_event_loop()

    async def start_scan(self, callback):
        """Start Scan with timeout"""
        raise NotImplementedError()

    async def stop_scan(self):
        """Stop Scan with timeout"""
        raise NotImplementedError()

    async def power_on(self):
        """Power on BLE Adapter"""
        raise NotImplementedError()

    async def power_off(self):
        """Power off BLE Adapter"""
        raise NotImplementedError()




import asyncio
from aioble.service import Service
from aioble.dotnet.characteristic import CharacteristicDotNet as Characteristic
from aioble.dotnet.utils import wrap_dotnet_task

from Windows.Devices.Bluetooth.GenericAttributeProfile import GattCommunicationStatus
from UWPBluetoothPython import UWPBluetooth

class ServiceDotNet(Service):
    """The Service DotNet Class"""
    def __init__(self, device, uuid, s_object, loop=None):
        super(ServiceDotNet, self).__init__(loop)
        self.loop = loop if loop else asyncio.get_event_loop()
        self.device = device
        self.s_object = s_object
        self.uuid = uuid
        self.characteristics = [] 
        self._dotnet_task = None
        self._uwp_bluetooth = UWPBluetooth()

    async def discover_characteristics(self):
        #print("Get Characteristics for {0}...".format(self.s_object.Uuid.ToString()))
        char_results = await wrap_dotnet_task(
            self._uwp_bluetooth.GetCharacteristicsAsync(self.s_object), loop=self.loop
        )

        if char_results.Status != GattCommunicationStatus.Success:
            print(
                "Could not fetch characteristics for {0}: {1}",
                self.s_object.Uuid.ToString(),
                char_results.Status,
            )
        else:
            self.characteristics = [Characteristic(self, characteristic) for characteristic in char_results.Characteristics]
        
        for characteristic in self.characteristics:
            await characteristic.discover_descriptors()


import CoreBluetooth
import Foundation
import asyncio
from typing import List

from aioble.service import Service
from aioble.corebluetooth.device import CoreBluetoothDevice
import aioble.corebluetooth.util as util

CoreBluetoothCharacteristic = None

class CoreBluetoothService(Service):
    def __init__(self, device : CoreBluetoothDevice, cbservice : CoreBluetooth.CBService, queue, *args, **kwargs):
        super(CoreBluetoothService, self).__init__(device, *args, **kwargs)
        self._cbservice = cbservice
        self._queue = queue

        self._identifier = self._cbservice.UUID().UUIDString()

        # deferred import is required due to to circular dependency between characteristic and service
        from aioble.corebluetooth.characteristic import CoreBluetoothCharacteristic as _characteristic
        global CoreBluetoothCharacteristic
        CoreBluetoothCharacteristic = _characteristic
        self._characteristics_by_identifier = None

        self._discover_characteristics_future = None

    @property
    def identifier(self):
        return self._identifier

    async def discover_characteristics(self) -> List[CoreBluetoothCharacteristic]:
        """Discover Service Characteristics"""
        if self._characteristics_by_identifier is None:
            self._discover_characteristics_future = asyncio.Future()
            await self._discover_characteristics()
            cbcharacteristics = await self._discover_characteristics_future
            self._characteristics_by_identifier = {cbchar.UUID().UUIDString(): CoreBluetoothCharacteristic(self, cbchar, self._queue) for cbchar in cbcharacteristics}
        return self._characteristics_by_identifier.values()

    async def characteristic_with_identifier(self, identifier : str) -> CoreBluetoothCharacteristic:
        await self.discover_characteristics()
        return self._characteristics_by_identifier.get(identifier)

    @util.dispatched_to_queue(wait=False)
    def _discover_characteristics(self):
        self._cbservice.peripheral().discoverCharacteristics_forService_(None, self._cbservice)

    def _did_discover_characteristics(self, characteristics : List[CoreBluetooth.CBCharacteristic], error : Foundation.NSError):
        if error is not None:
            self._discover_characteristics_future.set_exception(util.NSErrorException(error))
        else:
            self._discover_characteristics_future.set_result(characteristics)


import asyncio
from aioble.device import Device
import sys
import re
import dbussy as dbus
from dbussy import \
    DBUS

from aioble.bluezdbus.service import ServiceBlueZDbus as Service

_BLUEZ_DESTINATION = 'org.bluez'

_BLUEZ_OBJECT_PATH = '/org/bluez/hci0'

_DEVICE_INTERFACE = 'org.bluez.Device1'
_GATT_CHARACTERISTIC_INTERFACE = 'org.bluez.GattCharacteristic1'
_GATT_DESCRIPTOR_INTERFACE = 'org.bluez.GattDescriptor1'
_DBUS_OBJECT_MANAGER_INTERFACE = 'org.freedesktop.DBus.ObjectManager'

_CONNECT_METHOD = 'Connect'
_DISCONNECT_METHOD = 'Disconnect'
_START_NOTIFY_METHOD = 'StartNotify'
_STOP_NOTIFY_METHOD = 'StopNotify'
_WRITE_VALUE_METHOD = 'WriteValue'
_READ_VALUE_METHOD = 'ReadValue'
_GET_MANAGED_OBJECTS_METHOD = 'GetManagedObjects'

_device_path_regex_path = re.compile('^/org/bluez/hci0/dev((_[A-Z0-9]{2}){6})$')

class DeviceBlueZDbus(Device):
    """The Device Base Class"""

    def __init__(self, device, loop=None):
        super(DeviceBlueZDbus, self).__init__(loop)
        self.loop = loop if loop else asyncio.get_event_loop()
        self.address = '/org/bluez/%s/dev_%s' % ('hci0', device.replace(':', '_').upper())
        self._device_path = device
        self.properties = None
        self._notification_callbacks = {}
        self._dbus = None
        self.connect_succeeded = None
        self.disconnect_succeeded = None
        self.services_resolved = None
        self.is_services_resolved = False


    async def connect(self):
        """Connect to device"""

        try:
            # Connect to system bus
            self._dbus = await dbus.Connection.bus_get_async(DBUS.BUS_SYSTEM, private = False, loop = self.loop)
        except:
            print("ERROR: Invalid Device")

        # Define Signal Match
        self.properties_rule = {"type": "signal", "interface": "org.freedesktop.DBus.Properties", "member": "PropertiesChanged",
            "arg0": "org.bluez.Device1", "path": self._device_path}

        # Add Match Callback for PropertiesChanged
        await self._dbus.bus_add_match_action_async(self.properties_rule, self.properties_changed, None)

        # Assemble Connect Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = self._device_path,
            iface = _DEVICE_INTERFACE,
            method = _CONNECT_METHOD
        )

        # Connect
        await self._dbus.send_await_reply(message)

        while True:
            # Resolve Services if they haven't been so already
            services_resolved = await self.get_services_resolved()
            if services_resolved:
                break
            await asyncio.sleep(0.02, loop=self.loop)

        await self.discover_services()

    def _connect_succeeded(self):
        # Call User Callback
        if self.connect_succeeded != None:
            self.connect_succeeded()

    def _disconnect_succeeded(self, path):
        # Call User Callback
        if self.disconnect_succeeded != None:
            match = _device_path_regex_path.match(path)
            address = match.group(1)[1:].replace('_', ':').upper()
            #addres = '/org/bluez/%s/dev_%s' % ('hci0', path.replace(':', '_').upper())
            self.disconnect_succeeded(address)

    def _services_resolved(self):
        # Notify User that Services have been Discovered
        self.is_services_resolved = True
        if self.services_resolved != None:
            self.services_resolved()

    async def properties_changed(self, connection, message, data):
        """Properties Changed Signal"""
        if message.type == DBUS.MESSAGE_TYPE_SIGNAL:
            if message.member == "PropertiesChanged":
                if message.interface == "org.freedesktop.DBus.Properties":
                    if 'Connected' in list(message.objects)[1]:
                        if list(message.objects)[1]['Connected'][1]:
                            self._connect_succeeded()
                        else:
                            self._disconnect_succeeded(message.path)
                    if 'ServicesResolved' in list(message.objects)[1]:
                        if list(message.objects)[1]['ServicesResolved'][1]:
                            pass

            return DBUS.HANDLER_RESULT_HANDLED

    async def signal_parser(self, connection, message, data):
        """Characteristic Properties Changed Signal"""
        if message.type == DBUS.MESSAGE_TYPE_SIGNAL:
            if message.member == "PropertiesChanged":
                if message.interface == "org.freedesktop.DBus.Properties":
                    #print(self.address)
                    if message.path in self._notification_callbacks:
                        if 'Value' in list(message.objects)[1]:
                            # Call Callback with Data
                            self._notification_callbacks[message.path](message.path, list(message.objects)[1]['Value'][1])
                        elif 'Notifying' in list(message.objects)[1]:
                            pass
                            # Notifying Property Value
                            #print(list(message.objects)[1]['Notifying'][1])
            return DBUS.HANDLER_RESULT_HANDLED

    async def disconnect(self):
        """Disconnect to device"""

        # Remove Match Callback for PropertiesChanged
        await self._dbus.bus_remove_match_action_async(self.properties_rule, self.properties_changed, None)

        # Assemble Disonnect Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = dbus.valid_path(self._device_path),
            iface = _DEVICE_INTERFACE,
            method = _DISCONNECT_METHOD)

        await self._dbus.send_await_reply(message)

    async def is_connected(self):
        """Is Connected to device"""

        # Assemble IsConnected Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = dbus.valid_path(self._device_path),
            iface = DBUS.INTERFACE_PROPERTIES,
            method = "Get"
        )
        message.append_objects("s", dbus.valid_interface(_DEVICE_INTERFACE))
        message.append_objects("s", "Connected")
        reply = await self._dbus.send_await_reply(message)

        # Return Boolean IsConnected
        return reply.expect_return_objects("v")[0][1]

    async def read_char(self, uuid):
        """Read Service Char"""

        char_path = ""

        for s in self.services:
            for c in s.characteristics:
                if uuid == c.uuid:
                    char_path = c.path

        # Assemble Read Value Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = dbus.valid_path(char_path),
            iface = _GATT_CHARACTERISTIC_INTERFACE,
            method = _READ_VALUE_METHOD
        )
        message.append_objects("a{sv}", {})

        reply = await self._dbus.send_await_reply(message)
        values = reply.expect_return_objects("ay")[0]
        return values

    async def write_char(self, uuid, data):
        """Write Service Char"""

        char_path = ""

        for s in self.services:
            for c in s.characteristics:
                if uuid == c.uuid:
                    char_path = c.path

        bytes = [DBUS.subtype_byte(b) for b in data]

        # Assemble Write Value Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = dbus.valid_path(char_path),
            iface = _GATT_CHARACTERISTIC_INTERFACE,
            method = _WRITE_VALUE_METHOD
        )
        message.append_objects("ay", bytes)
        message.append_objects("a{sv}", {})

        try:
            reply = await self._dbus.send_await_reply(message)
            values = reply.expect_return_objects("")
        except dbus.DBusError:
            return ""

    async def start_notify(self, uuid, callback):
        """Start Notification Subscription"""

        char_path = ""

        for s in self.services:
            for c in s.characteristics:
                if uuid == c.uuid:
                    char_path = c.path

        # Add callback to active notify subscriptions
        self._notification_callbacks[char_path] = callback

        rule = {"type": "signal", "interface": "org.freedesktop.DBus.Properties", "member": "PropertiesChanged",
            "arg0": "org.bluez.GattCharacteristic1", "path": char_path}

        await self._dbus.bus_add_match_action_async(rule, self.signal_parser, None)

        # Assemble Start Notify Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = dbus.valid_path(char_path),
            iface = _GATT_CHARACTERISTIC_INTERFACE,
            method = _START_NOTIFY_METHOD
        )

        await self._dbus.send_await_reply(message)

    async def stop_notify(self, uuid):
        """Stop Notification Subscription"""

        char_path = ""

        for s in self.services:
            for c in s.characteristics:
                if uuid == c.uuid:
                    char_path = c.path

        # Remove callback to active notify subscriptions
        del self._notification_callbacks[char_path]

        rule = {"type": "signal", "interface": "org.freedesktop.DBus.Properties", "member": "PropertiesChanged",
            "arg0": "org.bluez.GattCharacteristic1", "path": char_path}

        # Remove Notify Callback
        await self._dbus.bus_remove_match_action_async(rule, self.signal_parser, None)

        # Assemble Stop Notify Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = dbus.valid_path(char_path),
            iface = _GATT_CHARACTERISTIC_INTERFACE,
            method = _STOP_NOTIFY_METHOD
        )

        await self._dbus.send_await_reply(message)

    async def get_services_resolved(self):
        """Is Services Resolved"""

        # Assemble Get Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = dbus.valid_path(self._device_path),
            iface = DBUS.INTERFACE_PROPERTIES,
            method = "Get"
        )
        message.append_objects("s", dbus.valid_interface(_DEVICE_INTERFACE))
        message.append_objects("s", 'ServicesResolved')
        reply = await self._dbus.send_await_reply(message)

        return reply.expect_return_objects("v")[0][1]

    async def discover_services(self):
        """Discover Device Services"""

        if self.services:
            return self.services
        else:
            print("Get Services...")
            message = dbus.Message.new_method_call \
            (
                destination= dbus.valid_bus_name(_BLUEZ_DESTINATION),
                path= dbus.valid_path("/"),
                iface= _DBUS_OBJECT_MANAGER_INTERFACE,
                method= _GET_MANAGED_OBJECTS_METHOD
            )
            # Dict of {Object Path, Dict of {String, Dict of {String, Variant}}} objects)
            reply = await self._dbus.send_await_reply(message)
            values = reply.expect_return_objects("a{oa{sa{sv}}}")[0]

            services_regex = re.compile(self._device_path + '/service[0-9abcdef]{4}$')

            self.services = [Service(
                device=self,
                path=objpath,
                uuid=objvalue['org.bluez.GattService1']['UUID'][1]) for objpath, objvalue in values.items() if
                services_regex.match(objpath)]

            for service in self.services:
                await service.resolve_characteristics()

            # Notify user
            self._services_resolved()

    async def _get_properties(self):
        """Get Properties"""

        # Assemble GetAll Properties Method Message
        message = dbus.Message.new_method_call \
        (
            destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
            path = dbus.valid_path(self._device_path),
            iface = DBUS.INTERFACE_PROPERTIES,
            method = "GetAll"
        )
        message.append_objects("s", dbus.valid_interface(_DEVICE_INTERFACE))
        reply = await self._dbus.send_await_reply(message)
        values = reply.expect_return_objects("a{sv}")[0]
        print(values)
        for propname in sorted(values.keys()) :
            proptype, propvalue = values[propname]
            sys.stdout.write("%s(%s) = %s\n" % (propname, proptype, repr(propvalue)))

    async def read_descriptor(self, uuid):
        """Read Characteristic Descriptor"""

        descriptor_path = ""

        for s in self.services:
            for c in s.characteristics:
                if uuid == c.uuid:
                    for d in c.descriptors:
                        descriptor_path = d.path

        try:
            # Assemble Read Value Method Message
            message = dbus.Message.new_method_call \
            (
                destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
                path = dbus.valid_path(descriptor_path),
                iface = _GATT_DESCRIPTOR_INTERFACE,
                method = _READ_VALUE_METHOD
            )
            message.append_objects("a{sv}", {})
        except Exception as ex:
            print("Exception, No Descriptor found for this Characteristic") 

        try:
            reply = await self._dbus.send_await_reply(message)
            values = reply.expect_return_objects("ay")[0]
            return values
        except:
            return []


import sys
import pathlib
import clr

_here = pathlib.Path(__file__).parent

# UWPBluetoothPython
sys.path.append(str(pathlib.Path(__file__).parent))
print(str(pathlib.Path(__file__).parent))
clr.AddReference("UWPBluetoothPython")



from aioble.centralmanager import CentralManager

from Windows.Devices.Bluetooth.Advertisement import BluetoothLEAdvertisementWatcher

class CentralManagerDotNet(CentralManager):
    """The Central Manager Dot Net Class"""
    def __init__(self, loop=None, **kwargs):
        super(CentralManagerDotNet, self).__init__(loop)
        self._device_found_callback = None
        self.devices = {}
        self._watcher = None
        self._device_found_callback = None
        self.service_uuids = None

    async def start_scan(self, callback, service_uuids=[]):
        self._device_found_callback = callback
        self.service_uuids = service_uuids
        self._watcher = BluetoothLEAdvertisementWatcher()
        self._watcher.Received += self._advertisement_received
        self._watcher.Start()

    def _advertisement_received(self, watcher, eventargs):
        if self._watcher == watcher:
            s_uuid = [str(i) for i in list(eventargs.Advertisement.ServiceUuids)]
            if self.service_uuids:
                if any(x in self.service_uuids for x in s_uuid):
                    if eventargs.BluetoothAddress not in self.devices:
                        self.devices[eventargs.BluetoothAddress] = {"Address", hex(eventargs.BluetoothAddress)}
                        s = (str(hex(eventargs.BluetoothAddress)))[2:].upper()
                        self._device_found_callback(eventargs.BluetoothAddress,
                                                    ':'.join(a + b for a, b in zip(s[::2], s[1::2])),
                                                    eventargs.Advertisement.LocalName)
            else:
                if eventargs.BluetoothAddress not in self.devices:
                    self.devices[eventargs.BluetoothAddress] = {"Address", hex(eventargs.BluetoothAddress)}
                    s = (str(hex(eventargs.BluetoothAddress)))[2:].upper()
                    self._device_found_callback(eventargs.BluetoothAddress,
                                                ':'.join(a+b for a,b in zip(s[::2], s[1::2])),
                                                eventargs.Advertisement.LocalName)

    async def stop_scan(self):
        """Stop Scan with timeout"""
        self._watcher.Stop()
#
    async def power_on(self):
        """Power on BLE Adapter"""
        raise NotImplementedError()

    async def power_off(self):
        """Power off BLE Adapter"""
        raise NotImplementedError()


import json
import re
import dbussy as dbus
from dbussy import \
    DBUS

from aioble.centralmanager import CentralManager

_BLUEZ_DESTINATION = 'org.bluez'

_BLUEZ_OBJECT_PATH = '/org/bluez/hci0'

_ADAPTER_INTERFACE = 'org.bluez.Adapter1'
_PROPERTIES_INTERFACE = 'org.freedesktop.DBus.Properties'

_INTROSPECT_METHOD = 'Introspect'
_START_DISCOVERY_METHOD = 'StartDiscovery'
_STOP_DISCOVERY_METHOD = 'StopDiscovery'
_SET_DISCOVERY_FILTER_METHOD = 'SetDiscoveryFilter'


class CentralManagerBlueZDbus(CentralManager):
    """The Central Manager Bluez Dbus Class"""

    def __init__(self, loop=None, **kwargs):
        super(CentralManagerBlueZDbus, self).__init__(loop)
        self.device = kwargs.get("device", "hci0")
        self._device_found_callback = None
        self.devices = {}
        self.devices_notified = {}
        self._dbus = None
        self._cached_devices = {}

    async def start_scan(self, callback, service_uuids=[]):
        # Set callback for new devices
        self._device_found_callback = callback
        self.service_uuids = service_uuids

        # Set device path regex
        self._device_path_regex = re.compile('^/org/bluez/' + 'hci0' + '/dev((_[A-Z0-9]{2}){6})$')

        # Connect to system bus
        self._dbus = await dbus.Connection.bus_get_async(DBUS.BUS_SYSTEM, private = False)

         # Perform Introspection
        message = dbus.Message.new_method_call(destination = _BLUEZ_DESTINATION, path = _BLUEZ_OBJECT_PATH, iface = DBUS.INTERFACE_INTROSPECTABLE, method = _INTROSPECT_METHOD)
        reply = await self._dbus.send_await_reply(message)
        introspection = dbus.Introspection.parse(reply.expect_return_objects("s")[0])

        # Assign Interfaces
        self._interfaces = introspection.interfaces_by_name
        self._adapter_interface = self._interfaces[_ADAPTER_INTERFACE]
        self._properties_interface = self._interfaces[_PROPERTIES_INTERFACE]

        # Assign Adapter Interface Methods
        self._adapter_interface_methods = self._adapter_interface.methods_by_name
        _set_discovery_filters_method = self._adapter_interface_methods[_SET_DISCOVERY_FILTER_METHOD]

        # Get method signature
        message = dbus.Message.new_method_call(destination = _BLUEZ_DESTINATION, path = _BLUEZ_OBJECT_PATH, iface = _ADAPTER_INTERFACE, method = _SET_DISCOVERY_FILTER_METHOD)
        sig = dbus.parse_signature(_set_discovery_filters_method.in_signature)

        discovery_filter = {'Transport': ['s', 'le']}

        # Call SetDiscovery Filter method with arguments
        message.append_objects("a{sv}", discovery_filter)

        #message.append_objects(sig, *argobjs)
        await self._dbus.send_await_reply(message)

        # Start Discovery
        self._dbus.add_filter(self.signal_parser, None)
        self._dbus.bus_add_match({"type": "signal", "interface": "org.freedesktop.DBus.ObjectManager",
                                  "member": "InterfacesAdded"})
        self._dbus.bus_add_match({"type": "signal", "interface": "org.freedesktop.DBus.Properties",
             "member": "PropertiesChanged","arg0": "org.bluez.Device1"})
        # Get method signature
        message = dbus.Message.new_method_call(destination = _BLUEZ_DESTINATION, path = _BLUEZ_OBJECT_PATH, iface = _ADAPTER_INTERFACE, method = _START_DISCOVERY_METHOD)

        await self._dbus.send_await_reply(message)

        # Add Devices that are already known
        await self._update_devices()

    async def stop_scan(self):
        """Stop Scan"""
        # Remove Signal Filter
        self._dbus.remove_filter(self.signal_parser, None)
        self._dbus.bus_remove_match({"type" : "signal", "interface" : "org.freedesktop.DBus.ObjectManager",
                                     "member" : "InterfacesAdded"})
        self._dbus.bus_remove_match({"type" : "signal", "interface" : "org.freedesktop.DBus.Properties",
                                     "member" : "PropertiesChanged", "arg0": "org.bluez.Device1"})
        # Get method signature
        message = dbus.Message.new_method_call(destination = _BLUEZ_DESTINATION, path = _BLUEZ_OBJECT_PATH, iface = _ADAPTER_INTERFACE, method = _STOP_DISCOVERY_METHOD)

        await self._dbus.send_await_reply(message)

    async def power_on(self):
        """Power on BLE Adapter"""
        raise NotImplementedError()

    async def power_off(self):
        """Power off BLE Adapter"""
        raise NotImplementedError()

    async def _update_devices(self):
        """Interface Added Signal"""
        message = dbus.Message.new_method_call(destination = _BLUEZ_DESTINATION, path = '/', iface = 'org.freedesktop.DBus.ObjectManager', method = 'GetManagedObjects')
        reply = await self._dbus.send_await_reply(message)
        for interfaces in reply.all_objects:
            for path in interfaces:
                if 'org.bluez.Device1' in interfaces[path]:
                    # Store Cached Device
                    self._cached_devices[path] = interfaces[path]["org.bluez.Device1"]

    def _search_for_uuids(self, path, properties):
        # Search for service UUIDs in device info
        for uuids in self.devices[path]["UUIDs"][1]:
            for s_uuid in self.service_uuids:
                if s_uuid in uuids:
                    # Notify user if they haven't been already
                    self._notify(path, properties)     
  
    def _add_new_device(self, path, properties):
        """Add New Device Found"""
        # Add new device
        self.devices[path] = properties

        # If searching by UUIDs
        if self.service_uuids:
            # If Interface was added during this scan
            if "UUIDs" in self.devices[path] and self.devices[path]["UUIDs"][1]:
                # Search for service UUIDs in device info
                self._search_for_uuids(path, properties)
            # If we don't have UUID information
            else:
                # Look if it is in cached devices
                if path in self._cached_devices:
                    # Add cached device info to device dict
                    self.devices[path] = {**self._cached_devices[path], **properties}

                    # Search for service UUIDs in device info
                    self._search_for_uuids(path, properties)
        else:
            # If Interface was added during this scan
            if "UUIDs" in self.devices[path] and self.devices[path]["UUIDs"][1]:
                # Notify user if they haven't been already
                self._notify(path, properties)
            # If we don't have UUID information
            else:
                # Look if it is in cached devices
                if path in self._cached_devices:
                    # Add cached device info to device dict
                    self.devices[path] = {**self._cached_devices[path], **properties}

                # Notify user if they haven't been already
                self._notify(path, properties)


    def _add_address(self, path, properties):
        # Add address field if missing
        if "Address" not in self.devices[path]:
            match = self._device_path_regex.match(path)
            address = match.group(1)[1:].replace('_', ':').lower()
            addressDict = {'Address': (type(address), address.upper())}
            properties = {**properties, **addressDict}
            # Update with Address field
            self.devices[path] = properties

    def _notify(self, path, properties):
        # If client hasn't been notified of new device
        if path not in self.devices_notified:
            # Add Address field from dbus path if no Address field exist
            self._add_address(path, properties)
            # Call Callback with new devices found
            if "Address" in self.devices[path] and "Alias" in self.devices[path]:
                self._device_found_callback(path, self.devices[path]["Address"][1], self.devices[path]["Alias"][1])
            elif "Address" in self.devices[path]:
                self._device_found_callback(path, self.devices[path]["Address"][1], "<unknown>")
            self.devices_notified[path] = properties

    def signal_parser(self, connection, message, data):
        """Interface Added Signal"""
        if message.type == DBUS.MESSAGE_TYPE_SIGNAL :
            if message.member == "InterfacesAdded":
                # Add New Device
                self._add_new_device(message.path, list(message.objects)[1]["org.bluez.Device1"])
                        
            elif message.member == 'PropertiesChanged':
                for e in message.objects:
                    if(e != "org.bluez.Device1"):
                        return DBUS.HANDLER_RESULT_HANDLED
                    else:
                        properties = list(message.objects)[1]
                        if message.path not in self.devices:
                            # Add New Device with new property
                            self._add_new_device(message.path, properties)
                        else:
                            # Update Existing Device with changed properties
                            self._add_new_device(message.path, {**self.devices[message.path], **properties})
            
        return DBUS.HANDLER_RESULT_HANDLED


import asyncio
import re
from aioble.characteristic import Characteristic
from aioble.bluezdbus.descriptor import DescriptorBlueZDbus as Descriptor 
import dbussy as dbus
from dbussy import \
    DBUS

_BLUEZ_DESTINATION = 'org.bluez'
_DBUS_OBJECT_MANAGER_INTERFACE = 'org.freedesktop.DBus.ObjectManager'
_GET_MANAGED_OBJECTS_METHOD = 'GetManagedObjects'

class CharacteristicBlueZDbus(Characteristic):
    """The Characteristic DotNet Class"""
    def __init__(self, service, path, uuid, loop=None):
        super(CharacteristicBlueZDbus, self).__init__(service)
        self.loop = loop if loop else asyncio.get_event_loop()
        self.service = service
        self.uuid = uuid
        self.path = path

    async def resolve_descriptors(self):
        # Connect to system bus
        self._dbus = await dbus.Connection.bus_get_async(DBUS.BUS_SYSTEM, private = False) 

        message = dbus.Message.new_method_call \
        (
            destination=_BLUEZ_DESTINATION,
            path="/",
            iface=_DBUS_OBJECT_MANAGER_INTERFACE,
            method=_GET_MANAGED_OBJECTS_METHOD
        )

        # Dict of {Object Path, Dict of {String, Dict of {String, Variant}}} objects)
        reply = await self._dbus.send_await_reply(message)
        values = reply.expect_return_objects("a{oa{sa{sv}}}")[0]

        descriptor_regex = re.compile(self.path + '/desc[0-9abcdef]{4}$')
        self.descriptors = [Descriptor(
            characteristic=self,
            path=objpath,
            uuid=objvalue['org.bluez.GattDescriptor1']['UUID'][1]) for objpath, objvalue in values.items() if
            descriptor_regex.match(objpath)]

        # print(self.uuid)
        # for d in self.descriptors: 
        #     try:
        #         print("d.path:")
        #         print(d.path)
        #     except:
        #         print("no descriptor")


import CoreBluetooth
import Foundation
import asyncio

import aioble.corebluetooth.util as util
from aioble.characteristic import Characteristic
from aioble.corebluetooth.service import CoreBluetoothService

class CoreBluetoothCharacteristic(Characteristic):
    """The CoreBluetooth implementation of a device characteristic"""
    def __init__(self, service : CoreBluetoothService, cbcharacteristic : CoreBluetooth.CBCharacteristic, queue, *args, **kwargs):
        super(CoreBluetoothCharacteristic, self).__init__(service, *args, **kwargs)
        self._cbcharacteristic = cbcharacteristic
        self._cbperipheral = cbcharacteristic.service().peripheral()
        self._queue = queue

        self._identifier = cbcharacteristic.UUID().UUIDString()

        self.read_value_future = None

    @property
    def identifier(self):
        return self._identifier

    async def read_value(self):
        if self._cbcharacteristic.properties() & CoreBluetooth.CBCharacteristicPropertyRead:
            self._read_value_future = asyncio.Future()
            await self._read_value()
            value = await self._read_value_future
            return value
        else:
            return None

    @util.dispatched_to_queue(wait=False)
    def _read_value(self):
        self._cbperipheral.readValueForCharacteristic_(self._cbcharacteristic)

    def _did_update_value(self, error : Foundation.NSError):
        future = self._read_value_future
        if future is not None:
            if error is not None:
                future.set_exception(util.NSErrorException(error))
            else:
                future.set_result(self._cbcharacteristic.value())

    async def read_value(self):
        if not self.can_read:
            self._read_value_future = asyncio.Future()
            await self._read_value()
            value = await self._read_value_future
            return value
        else:
            return None

    def can_broadcast(self):
        return self._cbcharacteristic.properties() & CoreBluetooth.CBCharacteristicPropertyBroadcast

    def can_read(self):
        return self._cbcharacteristic.properties() & CoreBluetooth.CBCharacteristicPropertyRead

    def can_write(self):
        return self._cbcharacteristic.properties() & CoreBluetooth.CBCharacteristicPropertyWrite

    def can_notify(self):
        return self._cbcharacteristic.properties() & CoreBluetooth.CBCharacteristicPropertyNotify

    def can_write_noreply(self):
        return self._cbcharacteristic.properties() & CoreBluetooth.CBCharacteristicPropertyWriteWithoutResponse

    @util.dispatched_to_queue(wait=False)
    def _read_value(self):
        self._cbperipheral.readValueForCharacteristic_(self._cbcharacteristic)

    def _did_update_value(self, error : Foundation.NSError):
        future = self._read_value_future
        if future is not None:
            if error is not None:
                future.set_exception(util.NSErrorException(error))
            else:
                future.set_result(self._cbcharacteristic.value())



import asyncio
from aioble.characteristic import Characteristic
from aioble.dotnet.descriptor import DescriptorDotNet as Descriptor 
from aioble.dotnet.utils import wrap_dotnet_task

from Windows.Devices.Bluetooth.GenericAttributeProfile import GattCommunicationStatus
from UWPBluetoothPython import UWPBluetooth

class CharacteristicDotNet(Characteristic):
    """The Characteristic DotNet Class"""
    def __init__(self, service, c_object, loop=None):
        super(CharacteristicDotNet, self).__init__(service)
        self.loop = loop if loop else asyncio.get_event_loop()
        self.service = service
        self.uuid = c_object.Uuid.ToString()
        self.c_object = c_object
        self._uwp_bluetooth = UWPBluetooth()
        self.descriptors = [] 

    async def discover_descriptors(self):
        #print("Get Characteristics for {0}...".format(self.s_object.Uuid.ToString()))
        descr_results = await wrap_dotnet_task(
            self._uwp_bluetooth.GetDescriptorsAsync(self.c_object), loop=self.loop
        )

        if descr_results.Status != GattCommunicationStatus.Success:
            print(
                "Could not fetch characteristics for {0}: {1}",
                self.c_object.Uuid.ToString(),
                descr_results.Status,
            )
        else:
            self.descriptors = [Descriptor(self, descriptor) for descriptor in descr_results.Descriptors]
            


import asyncio
from aioble.device import Device

from aioble.dotnet.utils import wrap_dotnet_task
from aioble.dotnet.service import ServiceDotNet as Service

from functools import wraps
from typing import Callable, Any

from UWPBluetoothPython import UWPBluetooth

from System import Array, Byte
from Windows.Devices.Bluetooth import BluetoothConnectionStatus
from Windows.Devices.Bluetooth.GenericAttributeProfile import (
    GattCharacteristic, GattCommunicationStatus
)

from Windows.Foundation import TypedEventHandler

class DeviceDotNet(Device):
    """The Device Base Class"""
    def __init__(self, address, loop=None):
        super(DeviceDotNet, self).__init__(loop)
        self.loop = loop if loop else asyncio.get_event_loop()
        self.address_hex = hex(address)
        self.address = address
        self.properties = None
        #UWP .NET
        self._dotnet_task = None
        self._uwp_bluetooth = UWPBluetooth()
        self._devices = {}
        self.services_resolved = None
        self.is_services_resolved = False
        self.connect_succeeded = None
        self.disconnect_succeeded = None
        self.services_resolved = None

    @property
    def name(self):
        raise NotImplementedError()

    async def connect(self):
        """Connect to device"""

        # Initiate Connection
        self._dotnet_task = await wrap_dotnet_task(
            self._uwp_bluetooth.FromBluetoothAddressAsync(self.address),
            loop=self.loop,
        )

        def _ConnectionStatusChanged_Handler(sender, args):
            print("_ConnectionStatusChanged_Handler: " + args.ToString())

        #Not Working for some reason?
        self._dotnet_task.ConnectionStatusChanged += _ConnectionStatusChanged_Handler

        #Discover Services
        await self.discover_services()

    async def disconnect(self):
        """Disconnect to device"""
        self._dotnet_task.Dispose()
        self._dotnet_task = None

    async def is_connected(self):
        if self._dotnet_task:
            return self._dotnet_task.ConnectionStatus == BluetoothConnectionStatus.Connected

        else:
            return False

    def _services_resolved(self):
        # Notify User that Services have been Discovered
        self.is_services_resolved = True
        if self.services_resolved != None:
            self.services_resolved()

    async def get_properties(self):
        """Get Device Properties"""
        raise NotImplementedError()

    async def discover_services(self):
        """Discover Device Services"""
        if self.services:
            return self.services
        else:
            #print("Get Services...")
            services = await wrap_dotnet_task(
                self._uwp_bluetooth.GetGattServicesAsync(self._dotnet_task), loop=self.loop
            )
            if services.Status == GattCommunicationStatus.Success:
                self.services = [Service(
                    device=s.Device,
                    uuid=s.Uuid.ToString(),
                    s_object=s) for s in services.Services]
            else:
                raise Exception("Could not get GATT services.")

            # Discover Characteristics for Each Service
            await asyncio.gather(
                *[
                    asyncio.ensure_future(service.discover_characteristics(), loop=self.loop)
                    for service in self.services
                ]
            )
            self._services_resolved()

    async def read_char(self, uuid):
        """Read Service Char"""
        for s_uuid, s in self.services.items():
            for c_uuid, c in s.characteristics.items():
                if str(uuid) == c_uuid:
                    characteristic = c.c_object

        if not characteristic:
            raise Exception("Characteristic {0} was not found!".format(uuid))

        read_results = await wrap_dotnet_task(
            self._uwp_bluetooth.ReadCharacteristicValueAsync(characteristic), loop=self.loop
        )
        status, value = read_results.Item1, bytearray(read_results.Item2)
        if status == GattCommunicationStatus.Success:
            #print("Read Characteristic {0} : {1}".format(uuid, value))
            pass
        else:
            raise Exception(
                "Could not read characteristic value for {0}: {1}",
                characteristic.Uuid.ToString(),
                status,
            )

        return list(value)

    async def write_char(self, uuid, data, response=False):
        """Write Service Char"""
        for s_uuid, s in self.services.items():
            for c_uuid, c in s.characteristics.items():
                if str(uuid) == c_uuid:
                    characteristic = c.c_object

        if not characteristic:
            raise Exception("Characteristic {0} was not found!".format(uuid))

        write_results = await wrap_dotnet_task(
            self._uwp_bluetooth.WriteCharacteristicValueAsync(
                characteristic, data, response
            ),
            loop=self.loop,
        )
        if write_results == GattCommunicationStatus.Success:
            #print("Write Characteristic {0} : {1}".format(uuid, data))
            pass
        else:
            raise Exception(
                "Could not write value {0} to characteristic {1}: {2}",
                data,
                characteristic.Uuid.ToString(),
                write_results,
            )

    async def start_notify(self, uuid, callback: Callable[[str, Any], Any], **kwargs):
        """Start Notification Subscription"""
        # Find the Characteristic object
        for s_uuid, s in self.services.items():
            for c_uuid, c in s.characteristics.items():
                if str(uuid) == c_uuid:
                    characteristic = c.c_object

        if self._notification_callbacks.get(str(uuid)):
            await self.stop_notify(uuid)

        dotnet_callback = TypedEventHandler[GattCharacteristic, Array[Byte]](
            _notification_wrapper(callback)
        )
        status = await wrap_dotnet_task(
            self._uwp_bluetooth.StartNotify(characteristic, dotnet_callback), loop=self.loop
        )
        if status != GattCommunicationStatus.Success:
            raise Exception(
                "Could not start notify on {0}: {1}",
                characteristic.Uuid.ToString(),
                status,
            )

    async def stop_notify(self, uuid):
        """Stop Notification Subscription"""
        # Find the Characteristic object
        for s_uuid, s in self.services.items():
            for c_uuid, c in s.characteristics.items():
                if str(uuid) == c_uuid:
                    characteristic = c.c_object
        # THIS IS NOT CURRENTLY STOPPING NOTIFICATION IMMEDIATELY
        status = await wrap_dotnet_task(
            self._uwp_bluetooth.StopNotify(characteristic), loop=self.loop
        )
        if status != GattCommunicationStatus.Success:
            raise Exception(
                "Could not start notify on {0}: {1}",
                characteristic.Uuid.ToString(),
                status,
            )

    async def read_descriptor(self, uuid):
        """Read Characteristic Descriptor"""
        for s_uuid, s in self.services.items():
            for c_uuid, c in s.characteristics.items():
                if str(uuid) == c_uuid:
                    characteristic = c.c_object

        if not characteristic:
            raise Exception("Characteristic {0} was not found!".format(uuid))

        read_results = await wrap_dotnet_task(
            self._uwp_bluetooth.ReadDescriptorValueAsync(characteristic), loop=self.loop
        )
        if read_results.Item2:
            status, value = read_results.Item1, bytearray(read_results.Item2)
            if status == GattCommunicationStatus.Success:
                #print("Read Characteristic {0} : {1}".format(uuid, value))
                pass
            else:
                raise Exception(
                    "Could not read descriptor value for {0}: {1}",
                    characteristic.Uuid.ToString(),
                    status,
                )
        else:
            return None

        return list(value)


def _notification_wrapper(func: Callable):
    @wraps(func)
    def dotnet_notification_parser(sender: Any, data: Any):
        return func(sender.Uuid.ToString(), bytearray(data))
    return dotnet_notification_parser



import asyncio
from aioble.descriptor import Descriptor 
from aioble.dotnet.utils import wrap_dotnet_task

from Windows.Devices.Bluetooth.GenericAttributeProfile import GattCommunicationStatus
from UWPBluetoothPython import UWPBluetooth

class DescriptorDotNet(Descriptor):
    """The Descriptor DotNet Class"""
    def __init__(self, characteristic, d_object, loop=None):
        super(DescriptorDotNet, self).__init__(characteristic)
        self.loop = loop if loop else asyncio.get_event_loop()
        self.characteristic = characteristic 
        self.uuid = d_object.Uuid.ToString()
        self.d_object = d_object
        self._uwp_bluetooth = UWPBluetooth()

    async def read_value(self):
        """Read Characteristic Descriptor"""
        #print("Get Characteristics for {0}...".format(self.s_object.Uuid.ToString()))
        if not self.characteristic:
            raise Exception("Characteristic {0} was not found!".format(uuid))

        read_results = await wrap_dotnet_task(
            self._uwp_bluetooth.ReadDescriptorValueAsync(self.characteristic.c_object), loop=self.loop
        )
        if read_results.Item2:
            status, value = read_results.Item1, bytearray(read_results.Item2)
            if status == GattCommunicationStatus.Success:
                #print("Read Characteristic {0} : {1}".format(uuid, value))
                pass
            else:
                raise Exception(
                    "Could not read descriptor value for {0}: {1}",
                    self.characteristic.Uuid.ToString(),
                    status,
                )
        else:
            return None

        return list(value)


import CoreBluetooth
import Foundation
import objc
import asyncio
from typing import List

import aioble.corebluetooth.util as util
from aioble.device import Device
from aioble.corebluetooth.centralmanager import CoreBluetoothCentralManager

CoreBluetoothService = None

class CoreBluetoothDevice(Device):
    """The CoreBluetooth concrete device implementation"""
    def __init__(self, manager: CoreBluetoothCentralManager, cbperipheral : CoreBluetooth.CBPeripheral, queue, *args, **kwargs):
        super(CoreBluetoothDevice, self).__init__(manager, *args, **kwargs)
        self._cbmanager = manager._cbmanager
        self._cbperipheral = cbperipheral
        self._cbperipheral.setDelegate_(self)
        self._queue = queue

        self._identifier = self._cbperipheral.identifier().UUIDString()

        # import of services is deferred due to circular dependency
        from aioble.corebluetooth.service import CoreBluetoothService as _service
        global CoreBluetoothService
        CoreBluetoothService = _service
        self._services_by_identifier = None

        self._did_connect_event = None
        self._did_disconnect_event = None
        self._discover_services_future = None

    # Public

    @property
    def identifier(self):
        return self._identifier

    @property
    def name(self):
        return self._cbperipheral.name()

    async def connect(self):
        """Connect to device"""
        if self._did_connect_event is None:
            self._did_connect_event = asyncio.Event()
        await self._connect()
        await self._did_connect_event.wait()
        self._did_connect_event = None

    async def disconnect(self):
        """Disconnect to device"""
        if self._did_disconnect_event is None:
            self._did_disconnect_event = asyncio.Event()
        await self._disconnect()
        await self._did_disconnect_event.wait()
        self._did_disconnect_event = None

    async def is_connected(self):
        """Is Connected to device"""
        return self._cbperipheral.state() is CoreBluetooth.CBPeripheralStateConnected

    async def get_properties(self):
        """Get Device Properties"""
        raise NotImplementedError()

    async def discover_services(self) -> List[CoreBluetoothService]:
        """Discover Device Services"""
        if self._services_by_identifier is None:
            self._discover_services_future = asyncio.Future()
            await self._discover_services()
            cbservices = await self._discover_services_future
            self._services_by_identifier = {cbservice.UUID().UUIDString(): CoreBluetoothService(self, cbservice, self._queue) for cbservice in cbservices}
        return self._services_by_identifier.values()

    async def service_with_identifier(self, identifier: str) -> CoreBluetoothService:
        await self.discover_services()
        return self._services_by_identifier.get(identifier)

    async def read_char(self):
        """Read Service Char"""
        raise NotImplementedError()

    async def write_char(self):
        """Write Service Char"""
        raise NotImplementedError()

    async def start_notify(self):
        """Start Notification Subscription"""
        raise NotImplementedError()

    async def stop_notify(self):
        """Stop Notification Subscription"""
        raise NotImplementedError()

    # Private

    @util.dispatched_to_queue(wait=False)
    def _connect(self):
        self._cbmanager.connectPeripheral_options_(self._cbperipheral, None)

    @util.dispatched_to_loop()
    async def _did_connect(self):
        self._did_connect_event.set()

    @util.dispatched_to_queue(wait=False)
    def _disconnect(self):
        self._cbmanager.cancelPeripheralConnection_(self._cbperipheral)

    @util.dispatched_to_loop()
    async def _did_disconnect(self):
        self._did_connect_event.set()

    @util.dispatched_to_queue(wait=False)
    def _discover_services(self):
        self._cbperipheral.discoverServices_(None)

    @util.dispatched_to_loop()
    async def _did_discover_services(self, services, error):
        if error is not None:
            self._discover_services_future.set_exception(util.NSErrorException(error))
        else:
            self._discover_services_future.set_result(services)

    @util.dispatched_to_loop()
    async def _did_discover_characteristics_for_service(self, cbservice : CoreBluetooth.CBService, cbcharacteristics : List[CoreBluetooth.CBCharacteristic], error : Foundation.NSError):
        service_identifier = cbservice.UUID().UUIDString()
        service = self._services_by_identifier.get(service_identifier)
        if service is not None:
            service._did_discover_characteristics(cbcharacteristics, error)

    @util.dispatched_to_loop()
    async def _did_update_value_for_characteristic(self, cbcharacteristic : CoreBluetooth.CBCharacteristic, error : Foundation.NSError):
        cbservice = cbcharacteristic.service()
        service = self._services_by_identifier.get(cbservice.UUID().UUIDString())
        if service is None:
            return

        char = await service.characteristic_with_identifier(cbcharacteristic.UUID().UUIDString())
        if char is None:
            return

        char._did_update_value(error)

    # CBPeripheralDelegate

    def peripheralDidUpdateName_(self, peripheral):
        pass

    def peripheral_didModifyServices_(self, peripheral, services):
        pass

    def peripheralDidUpdateRSSI_error_(self, peripheral, rssi, error):
        pass

    def peripheral_didReadRSSI_error_(self, peripheral, rssi, error):
        pass

    def peripheral_didDiscoverServices_(self, peripheral, error):
        self._did_discover_services(peripheral.services(), error)

    def peripheral_didDiscoverIncludedServices_error_(self, peripheral, service, error):
        pass

    def peripheral_didDiscoverCharacteristicsForService_error_(self, peripheral, service, error):
        self._did_discover_characteristics_for_service(service, service.characteristics(), error)

    def peripheral_didUpdateValueForCharacteristic_error_(self, peripheral, characteristic, error):
        self._did_update_value_for_characteristic(characteristic, error)

    def peripheral_didWriteValueForCharacteristic_error_(self, peripheral, characteristic, error):
        pass

    def peripheral_didUpdateNotificationStateForCharacteristic_error_(self, peripheral, characteristic, error):
        pass

    def peripheral_didDiscoverDescriptorsForCharacteristic_error_(self, peripheral, characteristic, error):
        pass

    def peripheral_didUpdateValueForDescriptor_error_(self, peripheral, descriptor, error):
        pass

    def peripheral_didWriteValueForDescriptor_error_(self, peripheral, descriptor, error):
        pass

    def peripheralIsReadyToSendWriteWithoutResponse_(self, peripheral):
        pass

    def peripheral_didOpenL2CAPChannel_error_(self, peripheral, channel, error):
        pass


import CoreBluetooth
import Foundation
import asyncio
import functools
import libdispatch
import objc

import aioble.corebluetooth.util as util
from aioble.centralmanager import CentralManager

CoreBluetoothDevice = None

class CoreBluetoothCentralManager(CentralManager):
    """Concrete implementation of the central manager protocol using CoreBluetooth API"""

    def __init__(self, loop=None, *args, **kwargs):
        super(CoreBluetoothCentralManager, self).__init__(loop, *args, **kwargs)
        self._lock = asyncio.Lock()
        self._queue = libdispatch.dispatch_queue_create(b'CoreBluetooth Queue', libdispatch.DISPATCH_QUEUE_SERIAL)
        self._cbmanager = CoreBluetooth.CBCentralManager.alloc().initWithDelegate_queue_options_(self, self._queue, None)
        self._queue_scan_count = 0
        self._queue_isScanning = False
        self._device_found_callback = None

        # We need to defer the import of CoreBluetoothDevice due to circular dependency
        from aioble.corebluetooth.device import CoreBluetoothDevice as _device
        global CoreBluetoothDevice
        CoreBluetoothDevice = _device
        self.devices = {}

    # Public

    async def start_scan(self, callback):
        async with self._lock:
            if self._device_found_callback is not None:
                raise "Can't scan while already scanning"
            self._device_found_callback = callback
            await self._increment_scan_count()

    async def stop_scan(self):
        async with self._lock:
            self._device_found_callback = None
            await self._decrement_scan_count()

    # Private

    @util.dispatched_to_queue()
    def _increment_scan_count(self):
        self._queue_scan_count = self._queue_scan_count + 1
        self._queue_update_scan_state_if_needed()

    @util.dispatched_to_queue()
    def _decrement_scan_count(self):
        if self._queue_scan_count > 0:
            self._queue_scan_count = self._queue_scan_count - 1
            self._queue_update_scan_state_if_needed()

    @util.dispatched_to_queue()
    def _update_scan_state_if_needed(self):
        self._queue_update_scan_state_if_needed()

    def _queue_update_scan_state_if_needed(self):
        if self._cbmanager.state() is not CoreBluetooth.CBManagerStatePoweredOn:
            return
        isScanning = self._cbmanager.isScanning()
        if self._queue_scan_count > 0 and not isScanning:
            self._cbmanager.scanForPeripheralsWithServices_options_(None, None)
        elif self._queue_scan_count == 0 and isScanning:
            self._cbmanager.stopScan()
        self._queue_isScanning = self._cbmanager.isScanning()

    @util.dispatched_to_loop()
    async def _notify_device_found(self, peripheral : CoreBluetooth.CBPeripheral):
        async with self._lock:
            if self._device_found_callback is not None:
                device = self.devices.get(peripheral.identifier().UUIDString())
                if device is None:
                    device = CoreBluetoothDevice(manager=self, cbperipheral=peripheral, queue=self._queue)
                    self.devices[device.identifier] = device
                    if asyncio.iscoroutinefunction(self._device_found_callback):
                        await self._device_found_callback(device)
                    else:
                        self._device_found_callback(device)

    @util.dispatched_to_loop()
    async def _peripheral_did_connect(self, peripheral : CoreBluetooth.CBPeripheral):
        device = self.devices.get(peripheral.identifier().UUIDString())
        if device:
            device._did_connect()

    @util.dispatched_to_loop()
    async def _peripheral_did_disconnect(self, peripheral : CoreBluetooth.CBPeripheral):
        device = self.devices.get(peripheral.identifier().UUIDString())
        if device:
            device._did_disconnect()

    # CBCentralManagerDelegate

    def centralManagerDidUpdateState_(self, manager):
        self._queue_update_scan_state_if_needed()

    def centralManager_didDiscoverPeripheral_advertisementData_RSSI_(self, manager, peripheral, data, rssi):
        self._notify_device_found(peripheral)

    def centralManager_didConnectPeripheral_(self, manager, peripheral):
        self._peripheral_did_connect(peripheral)

    def centralManager_didDisconnectPeripheral_(self, manager, peripheral):
        self._peripheral_did_disconnect(peripheral)



from aioble.descriptor import Descriptor 
import dbussy as dbus
from dbussy import \
    DBUS
    
_BLUEZ_DESTINATION = 'org.bluez'
_GATT_DESCRIPTOR_INTERFACE = 'org.bluez.GattDescriptor1'
_READ_VALUE_METHOD = 'ReadValue'

class DescriptorBlueZDbus(Descriptor):
    """The Characteristic DotNet Class"""
    def __init__(self, characteristic, path, uuid):
        super(DescriptorBlueZDbus, self).__init__(characteristic)
        self.characteristic = characteristic
        self.uuid = uuid
        self.path = path
        
    async def read_value(self):
        """Read Characteristic Descriptor"""
        # Connect to system bus
        self._dbus = await dbus.Connection.bus_get_async(DBUS.BUS_SYSTEM, private = False)

        try:
            # Assemble Read Value Method Message
            message = dbus.Message.new_method_call \
            (
                destination = dbus.valid_bus_name(_BLUEZ_DESTINATION),
                path = dbus.valid_path(self.path),
                iface = _GATT_DESCRIPTOR_INTERFACE,
                method = _READ_VALUE_METHOD
            )
            message.append_objects("a{sv}", {})
        except Exception as ex:
            print("Exception, Invalid Descriptor Path")

        try:
            reply = await self._dbus.send_await_reply(message)
            values = reply.expect_return_objects("ay")[0]
            return values
        except:
            return []


from distutils.core import setup
from setuptools import find_packages

setup(name='aioble',
      version='0.0.1',
      description='An asynchronous (asyncio) bluetooth low energy (BLE) python library',
      author='detect labs',
      author_email='code@detectlabs.io',
      url='https://github.com/detectlabs/aioble',
      download_url = 'https://github.com/detectlabs/aioble/archive/v0.0.1.tar.gz',
      packages=find_packages(exclude=('examples')),
      package_data={
          'aioble.dotnet': ['*.dll', ]
      },
      install_requires=[
          'dbussy; sys_platform == "linux"',
          'pyobjc; sys_platform == "darwin"',
          'pythonnet; sys_platform == "windows"'
      ]
)
